#include <stdio.h>

#include "aniDoc.h"
#include "aniUI.h"
#include "fltkCurveEditor.h"

#include "curveDisplay.h"

#include "Curve.h"
#include "CurveEditor.h"

#include "ctrlPanel.h"
#include "object.h"

#include "CurveEditor.h"

#include "Curves.h"

#include "CurveEvaluators.h"

#include "fltkCurveEditor.h"

extern CurveEditor *curve_editor;

#define LEFT		1
#define MIDDLE		2
#define RIGHT		3

#define LEFT_MOUSE_DOWN		1
#define LEFT_MOUSE_DRAG		2
#define LEFT_MOUSE_UP		3
#define MIDDLE_MOUSE_DOWN	4
#define MIDDLE_MOUSE_DRAG	5
#define MIDDLE_MOUSE_UP		6
#define RIGHT_MOUSE_DOWN	7
#define RIGHT_MOUSE_DRAG	8
#define RIGHT_MOUSE_UP		9
#define SHIFT_LEFT_DOWN		10
#define SHIFT_LEFT_DRAG		11
#define SHIFT_LEFT_UP		12
#define SHIFT_MIDDLE_DOWN	13
#define SHIFT_MIDDLE_DRAG	14
#define SHIFT_MIDDLE_UP		15
#define SHIFT_RIGHT_DOWN	16
#define SHIFT_RIGHT_DRAG	17
#define SHIFT_RIGHT_UP		18
#define CTRL_LEFT_DOWN		19
#define CTRL_LEFT_DRAG		20
#define CTRL_LEFT_UP		21
#define CTRL_MIDDLE_DOWN	22
#define CTRL_MIDDLE_DRAG	23
#define CTRL_MIDDLE_UP		24
#define CTRL_RIGHT_DOWN		25
#define CTRL_RIGHT_DRAG		26
#define CTRL_RIGHT_UP		27


extern CAniDoc *aniDoc;
extern CAniUI  *aniUI;
extern CurveEditor *curve_editor;
extern Object* objTable[];

extern float  getCurrentTime();
extern void   setCurrentTime(float);

extern void redrawScene();
extern void updateCurrentlySelectedCurve(const char*);
extern void updateMaxMinYValue(float, float);

// The following is not OOP at all
double minimum_time;
double time_interval;
double maximum_time;

double g_maximum_animation_time;
	
double minimum_value;
double value_interval;
double maximum_value;

double shift_x;
double shift_y;

double new_y_minimum;
double new_y_maximum;

int number_of_control_points=0;
int number_of_curve_segments=0;

//char g_currently_selected_name[128];

//----------------------------------------- help function ----------------------------------------------------------
void CurveEditor_UpdateVariables()
{
	double x_start;
	double x_division_interval;
	double x_end;

	double l_minimum_time = curve_editor->getMinimumTime();

	curve_editor->findNiceDivisions(l_minimum_time, curve_editor->getMaximumTime(), &x_start, &x_division_interval, &x_end);

	double l_shift_x = x_start - l_minimum_time;

	double y_start;
	double y_division_interval;
	double y_end;

	double l_minimum_value = curve_editor->getMinimumValue();

	curve_editor->findNiceDivisions(l_minimum_value, curve_editor->getMaximumValue(), &y_start, &y_division_interval, &y_end);

	double l_shift_y = y_start - l_minimum_value;

	double l_new_y_minimum = curve_editor->getMinimumValue();
	double l_new_y_maximum = curve_editor->getMaximumValue();


	Curve * currently_selected_curve = curve_editor->getCurrentlySelectedCurve();

	if (currently_selected_curve != NULL)
	{
		number_of_control_points = currently_selected_curve->getNumberOfControlPoints();

		number_of_curve_segments = currently_selected_curve->getNumberOfEvaluatedCurveSegments();
	}

	minimum_time=x_start;
	time_interval=x_division_interval;
	maximum_time=x_end;

	g_maximum_animation_time=x_end;
	
	minimum_value=y_start;
	value_interval=y_division_interval;
	maximum_value=y_end;

	shift_x=l_shift_x;
	shift_y=l_shift_y;

	new_y_minimum=l_new_y_minimum;
	new_y_maximum=l_new_y_maximum;
}

void reevaluate()
{
	// re-evaluate all curves
	Curves * curves = Curves::getInstance();
	double current_time=getCurrentTime();

	for (Curves::const_iterator curve_iterator = curves->begin(); curve_iterator != curves->end(); ++curve_iterator)
	{
		double current_value = curve_iterator->evaluateCurveAt(current_time);

		// for multiple objects
		// from curve name, set the value for the corresponding widget
		// name in the format curve_name#obj_id@widget_id
		char buf[64];
		int obj_id, id;

		strcpy(buf, curve_iterator->getName().c_str());
		strtok(buf, "#@");
		obj_id=atoi(strtok(NULL, "#@"));
		id=atoi(strtok(NULL, "#@"));

		objTable[obj_id]->setCurveValue(id, (float)current_value);
	}
}
//---------------------------------------- member functions ------------------------------------------
// menu definition

curveDisplay::curveDisplay(int x, int y, int w, int h, const char *l, fltkCurveEditor *f) : Fl_Gl_Window(x,y,w,h,l)
{
	isAnEvent=0;

	// create popup menu
	Fl_Menu_Button *m_pPopupMenu = new Fl_Menu_Button(0,0,600,400,"popup");
	m_pPopupMenu->type(Fl_Menu_Button::POPUP3);

	menutable = new Fl_Menu_Item[128];	// hopefully, 128 should be enough
	memset(menutable, 0, 128*sizeof(Fl_Menu_Item));

	// add curve types
	menutable[0].label( "Curve types" );
	menutable[0].shortcut(0);
	menutable[0].callback((Fl_Callback*)0);
	menutable[0].user_data((void*)0);
	menutable[0].flags=FL_SUBMENU;
	
	CurveEvaluators * curve_evaluators = CurveEvaluators::getInstance();
	CurveEvaluators::const_iterator j;
	int i=1;
	for( j = curve_evaluators->begin(); j != curve_evaluators->end(); ++j ) {
		menutable[i].label( (*j).first.c_str() );
		menutable[i].shortcut(0);
		menutable[i].callback(cb_set_curve_type);
		menutable[i].user_data( (void *)((*j).first.c_str()) );
		i++;
	}
	menutable[i].label( 0 );
	i++;

	menutable[i].label( "Animation Variables" );
	menutable[i].shortcut(0);
	menutable[i].callback((Fl_Callback*)0);
	menutable[i].user_data((void*)0);
	menutable[i].flags=FL_SUBMENU;
	i++;

	menutable[i].label( 0 );
	m_nMenuStart=m_nMenuEnd=i;

	m_pPopupMenu->menu(menutable);

	father=f;
}

curveDisplay::~curveDisplay()
{
	delete m_pPopupMenu;
	delete menutable;
}

void curveDisplay::resetCurveMenu(Object *obj) 
{ 
	ctrlPanel::map_iterator j;
	int i=m_nMenuStart;

	for( j = obj->m_pPanel->begin(); j != obj->m_pPanel->end(); ++j ) {
		if ((*j).second->curve && 
				(strcmp((*j).second->name, "Quaternion X") != 0) &&
				(strcmp((*j).second->name, "Quaternion Y") != 0) &&
				(strcmp((*j).second->name, "Quaternion Z") != 0) &&
				(strcmp((*j).second->name, "Quaternion W") != 0) &&
				(strcmp((*j).second->name, "Translate X") != 0) &&
				(strcmp((*j).second->name, "Translate Y") != 0) &&
				(strcmp((*j).second->name, "Translate Z") != 0) &&
				(strcmp((*j).second->name, "Zoom") != 0) )  {
			menutable[i].label( (*j).second->name );
			menutable[i].shortcut(0);
			menutable[i].callback((Fl_Callback*)cb_toggle_curve);
			menutable[i].user_data((void *)((*j).first));
			menutable[i].flags=FL_MENU_TOGGLE;
			if ((*j).second->toggle) menutable[i].set();
			i++;
		}
	}
	menutable[i].label( 0 );
	if (i<m_nMenuEnd) memset(menutable+i, 0, (m_nMenuEnd-i)*sizeof(Fl_Menu_Item));
	m_nMenuEnd=i;
}

void curveDisplay::cb_set_curve_type(Fl_Widget* o, void* v)
{
	Curve * currently_selected_curve = curve_editor->getCurrentlySelectedCurve();
		
	if (currently_selected_curve != NULL)
	{
		currently_selected_curve->setCurveType((char*)v);
	}

	CurveEditor_UpdateVariables();

	aniUI->m_pCurveEditor->refresh();
}

void curveDisplay::cb_toggle_curve(Fl_Widget* o, void* v)
{
	Object *currentObject=aniDoc->getScene()->m_pActiveObj;
	int obj_id=currentObject->getObjID();
	int id=(int)v;
		
	char buf[64];
	sprintf(buf, "%s#%d@%d", currentObject->getCurveName(id), obj_id, id);
	currentObject->toggleCurve(id);
	curve_editor->toggleShowAnimationCurve(buf);
	CurveEditor_UpdateVariables();
	aniUI->m_pCurveEditor->updateValueScale();
	updateMaxMinYValue(new_y_maximum, new_y_minimum);
	aniUI->m_pCurveEditor->redraw();
#ifndef WIN32
	aniUI->m_pCurveEditor->refresh();
#endif
}


void curveDisplay::refresh() 
{ 
	reevaluate();
	redraw();
}

void curveDisplay::draw()
{
	if(!valid())
	{
		glClearColor(0.7f, 0.7f, 0.7f, 1.0);

		// We're only using 2-D, so turn off depth 
		glDisable( GL_DEPTH_TEST );

		ortho();
	}

	glClear( GL_COLOR_BUFFER_BIT);
	
	if (isAnEvent) {
		isAnEvent=0;	// clear it after processing
		double t;

		switch (eventToDo) {
			case LEFT_MOUSE_DOWN:
					curve_editor->mouseDown(coordx, coordy);
					CurveEditor_UpdateVariables();
				break;
			case LEFT_MOUSE_DRAG:
					curve_editor->mouseMove(coordx, coordy);
				break;
			case LEFT_MOUSE_UP:
					curve_editor->mouseUp(coordx, coordy);
					CurveEditor_UpdateVariables();
				break;
			case CTRL_LEFT_DOWN:
					curve_editor->controlClick(coordx, coordy);
					CurveEditor_UpdateVariables();
				break;
			case MIDDLE_MOUSE_DOWN:
					curve_editor->altClick(coordx, coordy);
					CurveEditor_UpdateVariables();
					father->updateValueScale();
					father->updateMaxMin(new_y_maximum, new_y_minimum);
					father->redraw();
				break;
			case SHIFT_LEFT_DOWN:
					curve_editor->shiftMouseDown(coordx, coordy);				
				break;
			case SHIFT_LEFT_DRAG:
					curve_editor->shiftMouseDrag(coordx, coordy);				
				break;
			case SHIFT_LEFT_UP:
					curve_editor->shiftMouseUp(coordx, coordy);
					CurveEditor_UpdateVariables();
					father->updateTimeScale();
					father->updateValueScale();
					father->updateMaxMin(new_y_maximum, new_y_minimum);
					father->redraw();
				break;
			case SHIFT_MIDDLE_DOWN:
				t=curve_editor->getTime(coordx);
				setCurrentTime(t);
				break;
			case SHIFT_MIDDLE_DRAG:
				t=curve_editor->getTime(coordx);
				setCurrentTime(t);
				break;
			case SHIFT_MIDDLE_UP:
				t=curve_editor->getTime(coordx);
				setCurrentTime(t);
				break;
			default:
				break;
		}
	}

	// redraw the curve editor
	curve_editor->drawCurveEditor(getCurrentTime());
	reevaluate();
	redrawScene();

	// set currently selected curve name
	Curve *c=curve_editor->getCurrentlySelectedCurve();
	if (c!=NULL) {
		updateCurrentlySelectedCurve(c->getName().c_str());
	} else 
		updateCurrentlySelectedCurve("");

}

int curveDisplay::handle(int event)
{
	switch(event)
	{
	case FL_PUSH:
		coordx = Fl::event_x();
		coordy = Fl::event_y();
		switch (Fl::event_button()) {
		case LEFT:
			if (!Fl::event_state(FL_ALT)) {
				if (Fl::event_state(FL_SHIFT)) eventToDo=SHIFT_LEFT_DOWN;
				else if (Fl::event_state(FL_CTRL)) eventToDo=CTRL_LEFT_DOWN;
				else eventToDo=LEFT_MOUSE_DOWN;
				break;
			}	// ALT+LEFT==MIDDLE
		case MIDDLE:
			if (Fl::event_state(FL_SHIFT)) eventToDo=SHIFT_MIDDLE_DOWN;
			else if (Fl::event_state(FL_CTRL)) eventToDo=CTRL_MIDDLE_DOWN;
			else eventToDo=MIDDLE_MOUSE_DOWN;
			break;
		case RIGHT:
			if (Fl::event_state(FL_SHIFT)) eventToDo=SHIFT_RIGHT_DOWN;
			else if (Fl::event_state(FL_CTRL)) eventToDo=CTRL_RIGHT_DOWN;
			else eventToDo=RIGHT_MOUSE_DOWN;
			break;
		}
		isAnEvent=1;
		redraw();
		break;
	case FL_DRAG:
		coordx = Fl::event_x();
		coordy = Fl::event_y();
		switch (Fl::event_button()) {
		case LEFT:
			if (!Fl::event_state(FL_ALT)) {
				if (Fl::event_state(FL_SHIFT)) eventToDo=SHIFT_LEFT_DRAG;
				else if (Fl::event_state(FL_CTRL)) eventToDo=CTRL_LEFT_DRAG;
				else eventToDo=LEFT_MOUSE_DRAG;
				break;
			}	// ALT+LEFT==MIDDLE
		case MIDDLE:
				if (Fl::event_state(FL_SHIFT)) eventToDo=SHIFT_MIDDLE_DRAG;
				else if (Fl::event_state(FL_CTRL)) eventToDo=CTRL_MIDDLE_DRAG;
				else eventToDo=MIDDLE_MOUSE_DRAG;
				break;
		case RIGHT:
			if (Fl::event_state(FL_SHIFT)) eventToDo=SHIFT_RIGHT_DRAG;
			else if (Fl::event_state(FL_CTRL)) eventToDo=CTRL_RIGHT_DRAG;
			else eventToDo=RIGHT_MOUSE_DRAG;
			break;
		}
		isAnEvent=1;
		redraw();
		break;
	case FL_RELEASE:
		coordx = Fl::event_x();
		coordy = Fl::event_y();
		switch (Fl::event_button()) {
		case LEFT:
			if (!Fl::event_state(FL_ALT)) {
				if (Fl::event_state(FL_SHIFT)) eventToDo=SHIFT_LEFT_UP;
				else if (Fl::event_state(FL_CTRL)) eventToDo=CTRL_LEFT_UP;
				else eventToDo=LEFT_MOUSE_UP;
				break;
			}	// ALT+LEFT==MIDDLE
		case MIDDLE:
			if (Fl::event_state(FL_SHIFT)) eventToDo=SHIFT_MIDDLE_UP;
			else if (Fl::event_state(FL_CTRL)) eventToDo=CTRL_MIDDLE_UP;
			else eventToDo=MIDDLE_MOUSE_UP;
			break;
		case RIGHT:
			if (Fl::event_state(FL_SHIFT)) eventToDo=SHIFT_RIGHT_UP;
			else if (Fl::event_state(FL_CTRL)) eventToDo=CTRL_RIGHT_UP;
			else eventToDo=RIGHT_MOUSE_UP;
			break;
		}
		isAnEvent=1;
		redraw();
	default:
		return 0;
		break;
	}
	return 1;
}

