// 
// scene.cpp
// 

#include <stdio.h>

#include "scene.h"
#include "aniDoc.h"
#include "aniUI.h"
#include "modeler.h"
#include "ctrlPanel.h"
#include <math.h>
#include "Curves.h"

#include "fltkCurveEditor.h"

extern CAniUI *aniUI;

extern char *g_rayfilename;
#define M_PI 3.1415926

extern ctrlPanel* g_panel;

const Vec3 XAXIS = {1,0,0};
const Vec3 YAXIS = {0,1,0};
const Vec3 ZAXIS = {0,0,1};


void Scene::draw_circle(double x, double y, double radius)
{   
   glPushMatrix();

   glTranslated(x, y, 0);
   glScaled(radius, radius, 1.);
   
   glBegin(GL_LINE_LOOP);
   const double segs = 60;
   for (int i = 0; i < segs; i++) {
	   glVertex2d((cos(i * 2 * M_PI / segs)), 
		   (sin(i * 2 * M_PI / segs)));
   }
   glEnd();
   
   glPopMatrix();
}

Scene::Scene(Object* obj) : objects() 
{
	m_nObjCnt=0;
	m_pView=NULL;
	m_pObjList=NULL;
	m_pActiveObj=NULL;
	add( obj );
	center.x = 0;center.y = 0;center.z = 0;
	orig_center.x = 0;	orig_center.y = 0;	orig_center.z = 0;
	radius = 200.0;
	vfrom.x = vfrom.y = vfrom.z = 0;
	vto.x = vto.y = vto.z = 0;
	dragging = viewpointEdit = false;
	qdown.x = qdown.y = qdown.z = 0;
	qnow.x = qnow.y = qnow.z = 0;
	qdown.w = 1; qnow.w = 1;
	model[0][0] = 1;	model[1][0] = 0;	model[2][0] = 0;	model[3][0] = 0;
	model[0][1] = 0;	model[1][1] = 1;	model[2][1] = 0;	model[3][1] = 0;
	model[0][2] = 0;	model[1][2] = 0;	model[2][2] = 1;	model[3][2] = 0;
	model[0][3] = 0;	model[1][3] = 0;	model[2][3] = 0;	model[3][3] = 1;
	zoom = orig_zoom = -20;
	current=NONE;
	g_panel->curve(qnowxid,"Quaternion X",-5,5,0);
	Curves::getInstance()->addCurve(qnowx, -5, 5,0);
	g_panel->curve(qnowyid,"Quaternion Y",-5,5,0);
	Curves::getInstance()->addCurve(qnowy, -5, 5,0);
	g_panel->curve(qnowzid,"Quaternion Z",-5,5,0);
	Curves::getInstance()->addCurve(qnowz, -5, 5,0);
	g_panel->curve(qnowwid,"Quaternion W",-5,5,1);
	Curves::getInstance()->addCurve(qnoww, -5, 5,1);
	g_panel->curve(centerxid,"Translate X",-100,100,0);
	Curves::getInstance()->addCurve(centerx, -100, 100,0);
	g_panel->curve(centeryid,"Translate Y",-100,100,0);
	Curves::getInstance()->addCurve(centery, -100, 100,0);
	g_panel->curve(centerzid,"Translate Z",-100,100,0);
	Curves::getInstance()->addCurve(centerz, -100, 100,0);
	g_panel->curve(zoomid,"Zoom",-100,100,-20);
	Curves::getInstance()->addCurve(zoomed, -100, 100,-20);
	g_panel->curve(viewpointid, "Viewpoint", -1,1,0);
	Curves::getInstance()->addCurve(viewpoint, -1,1,0);

}

Scene::~Scene()
{
	if (m_pObjList) {
		m_pObjList->hide();
		delete m_pObjList;
	}
	if (m_pView) {
		m_pView->hide();
		aniUI->removeView(m_pView);
		delete m_pView;
	}

	for( giter g = objects.begin(); g != objects.end(); ++g ) {
		delete (*g);
	}
}

void Scene::createView()
{
	if (m_pView!=NULL) delete m_pView;

	// create a view for it 
	m_pView = new CGLView(600, 100, 400, 400, "Scene", this);
	m_pView->setUI(m_pDoc->getUI());
	m_pView->resizable(m_pView);
}

void Scene::createObjList()
{
	if (m_pObjList!=NULL) delete m_pObjList;

	// create the object list panel for the scene
	int i=0;
	giter g;

	m_pObjList=new Fl_Window(100, 100, 200, 20*(m_nObjCnt+1)+10, "Object List");

		Fl_Box *box=new Fl_Box(5, 5, 190, 18, "             Object                V   A");

		for( g = objects.begin(); g != objects.end(); ++g ) {
			i++;
			// create buttons for object
			char *buf=new char[16];

			sprintf(buf, "%s<%d>", (*g)->getObjName(), (*g)->getObjID());
			Fl_Button *b1=new Fl_Button(10, i*20+5, 150, 18, buf);
			b1->user_data(new scene_params((void*)this,(Object*)(*g)));
			b1->callback(cb_fork_object);

			Fl_Check_Button *b2=new Fl_Check_Button(160, i*20+5, 18, 18, "");
			b2->value(1);
			b2->user_data(new scene_params((void*)this,(Object*)(*g)));
			b2->callback(cb_toggle_object);
		}

		i=0;
		Fl_Group* o=new Fl_Group(180, 20, 20, 20*(m_nObjCnt));
		for( g = objects.begin(); g != objects.end(); ++g ) {
			i++;
			Fl_Check_Button *b3=new Fl_Check_Button(180, i*20+5, 18, 18, "");
			b3->type(102);
			b3->user_data(new scene_params((void*)this,(Object*)(*g)));
			b3->callback(cb_activate_object);
			if (i==1) b3->value(1);
		}
		o->end();

	m_pObjList->end();

	m_pActiveObj=objects.front();
	aniUI->m_pCurveEditor->resetCurveMenu(m_pActiveObj);
}

void Scene::saveImage(char *iname) 
{ 
	m_pView->saveImage(iname); 
}

void Scene::show()
{
	if (m_pView) m_pView->show();

/* we only support one object now */
//	if (m_pObjList) m_pObjList->show();
	// if the object doesn't currently have a subwindow, show the panel and create a scene for the object
	if (!(m_pActiveObj->m_pPanel->visible())) {
		m_pActiveObj->m_pPanel->show();
		aniUI->addPanel(m_pActiveObj->m_pPanel);
	}

	// open control depending on the mode
	if (g_mode==CURVE_MODE) {

		// show fltkCurveEditor
		aniUI->m_pCurveEditor->show();
		aniUI->m_pCurveEditor->setScene(aniUI->getDocument()->getScene());
	
		// hide the object panel
		if (m_pDoc->getScene()) {
			m_pDoc->getScene()->m_pActiveObj->m_pPanel->hide();
		}

		aniUI->refreshAll();
	} else {
		// hide fltkCurveEditor
		aniUI->m_pCurveEditor->hide();

		// if the object doesn't currently have a subwindow, show the panel and create a scene for the object
		if (m_pDoc->getScene()) {
			m_pDoc->getScene()->m_pActiveObj->m_pPanel->show();
			aniUI->addPanel(m_pDoc->getScene()->m_pActiveObj->m_pPanel);
		}

		aniUI->refreshAll();
	}

	aniUI->addView(m_pView);
}

void Scene::redraw(int drawMode)
{
	Quat qdrag;
	draw_mode_setup(drawMode);
	glLoadIdentity();
	if(viewpointEdit)
	{
		glTranslated(0,0,zoom);
		if(dragging && current == ROTATE)
		{
			qdrag = Qt_FromBallPoints(vfrom, vto);
			if(qdrag.x || qdrag.y || qdrag.z || qdrag.w)
			{
				qnow =Qt_Mul(qdrag, qdown);
			}
		}
		else if(dragging && (current == CONSTRAIN_ABOUT_X ||
							 current == CONSTRAIN_ABOUT_Y ||
							 current == CONSTRAIN_ABOUT_Z))
		{
			switch(current)
			{
			case CONSTRAIN_ABOUT_X:
				vfrom = ConstrainToAxis(vfrom, XAXIS); 
				vto = ConstrainToAxis(vto, XAXIS);
				break;
			case CONSTRAIN_ABOUT_Y:
				vfrom = ConstrainToAxis(vfrom, YAXIS); 
				vto = ConstrainToAxis(vto, YAXIS);
				break;
			case CONSTRAIN_ABOUT_Z:
				vfrom = ConstrainToAxis(vfrom, ZAXIS); 
				vto = ConstrainToAxis(vto, ZAXIS);
				break;
			}
			qdrag = Qt_FromBallPoints(vfrom, vto);
			if(qdrag.x || qdrag.y || qdrag.z || qdrag.w)
			{
				qnow =Qt_Mul(qdrag, qdown);
			}

		}
		Qt_ToMatrix(Qt_Conj(qnow),model);
		glMultMatrixd((double*)model);

		glTranslated(-center.x,-center.y,-center.z);
	}
	else
	{
		zoom = g_panel->getControlValue(zoomid);
		glTranslated(0,0,zoom);
		qnow.x = g_panel->getControlValue(qnowxid);
		qnow.y = g_panel->getControlValue(qnowyid);
		qnow.z = g_panel->getControlValue(qnowzid);
		qnow.w = g_panel->getControlValue(qnowwid);
		Qt_ToMatrix(Qt_Conj(qnow),model);
		glMultMatrixd((double*)model);
		center.x = g_panel->getControlValue(centerxid);
		center.y = g_panel->getControlValue(centeryid);
		center.z = g_panel->getControlValue(centerzid);
		glTranslated(-center.x, -center.y, -center.z);
	}

	glPushMatrix();
	for( giter g = objects.begin(); g != objects.end(); ++g ) {
		if ((*g)->isVisible())
			(*g)->redraw((*g)->getObjID(),drawMode);
	}
	glPopMatrix();
	if(viewpointEdit && dragging)
	{   
	   glPushAttrib(GL_LINE_BIT | GL_LIGHTING_BIT);
	   glDisable(GL_LIGHTING);
		glPushMatrix();
			glTranslated(center.x,center.y,center.z);
			  glLineWidth(2);
			  glBegin(GL_LINES);
			  glColor3f(1,0,0); glVertex3f(-1,0,0); 
			  glColor3f(1,.2,.2); glVertex3f(1,0,0);
			  glColor3f(0,.4,0); glVertex3f(0,-1,0); 
			  glColor3f(0,1,0); glVertex3f(0,1,0);
			  glColor3f(0,0,1); glVertex3f(0,0,-1); 
			  glColor3f(.2,.2,1); glVertex3f(0,0,1);
			  glEnd();
	   glPopMatrix();

		glColor3f(1,1,0);
		glMatrixMode( GL_PROJECTION );
		glPushMatrix();
			glLoadIdentity();
			glOrtho(-m_pView->w()*.5, m_pView->w()*.5, -m_pView->h()*.5, m_pView->h()*.5, -1, 1);

			glMatrixMode( GL_MODELVIEW );
			glPushMatrix();
				glLoadIdentity();

				glLineWidth(3);
				draw_circle(0,0, radius);

				glMatrixMode( GL_PROJECTION );
			glPopMatrix();

			glMatrixMode( GL_MODELVIEW );
		glPopMatrix();
	   glPopAttrib();

	}
	draw_mode_finish(drawMode);

}

void Scene::refresh()
{
	m_pView->refresh();
}

/* draw_mode_setup
 *
 * models should call this after clearing the window in their redraw
 * function.  it sets up default material and light properties, and
 * does initialization specific to the drawing mode.
 */

void Scene::draw_mode_setup(int drawMode)
{
    if ( drawMode & MODE_RAYTRACE ) {
		setDrawMode( drawMode, g_rayfilename );
		} 
	else {
		setDrawMode( drawMode, NULL );
		}

    GLfloat light_position[] = { 10.0f, 10.0f, 10.0f, 0.0f };

	GLfloat mat_ambient[4];
	GLfloat mat_diffuse[4];
	GLfloat mat_specular[4];
	GLfloat mat_shininess;

    mat_ambient[0] = mat_ambient[1] = mat_ambient[2] = 0.5;
    mat_ambient[3] = 1.0;
    mat_diffuse[0] = mat_diffuse[1] = mat_diffuse[2] = 0.5;
    mat_diffuse[3] = 1.0;
    mat_specular[0] = mat_specular[1] = mat_specular[2] = mat_specular[3] = 1.0;
    mat_shininess = 100.0;

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    
    glEnable( GL_DEPTH_TEST );

    if ( drawMode & MODE_NORMAL )
    {
		/* initialization for "normal" mode. */
	
		glShadeModel( GL_SMOOTH );
		glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_ambient );
		glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular );
		glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, &mat_shininess );
		glLightfv( GL_LIGHT0, GL_POSITION, light_position );

		glEnable( GL_LIGHTING );
		glEnable( GL_LIGHT0 );
		glEnable( GL_NORMALIZE );
	
		glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE );
		glPolygonMode( GL_FRONT, GL_FILL );
		glCullFace( GL_BACK );
		glEnable( GL_CULL_FACE ); 
    }

    if ( drawMode & MODE_WIREFRAME )
    {
		/* initialization for "wireframe" mode. */
	
		glDisable( GL_LIGHTING );
		glDisable( GL_NORMALIZE );
	
		glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );

		glColor3f( 0.5, 0.5, 0.5 );
    }

    if ( drawMode & MODE_FLATSHADE )
    {
		/* initialization for "flat shaded" mode. */
	
		glDisable( GL_LIGHTING );
		glDisable( GL_NORMALIZE );
	
		glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	
		glColor3f( 0.5, 0.5, 0.5 );
    }
}

void Scene::draw_mode_finish( int drawMode )
{

    if ( drawMode & MODE_RAYTRACE )		// ask modeler.dll to close the ray file
    {
		setDrawMode( drawMode, NULL );
    }

}

void Scene::cb_fork_object(Fl_Widget* o, void* v)
{
	scene_params* args=(scene_params *)(o->user_data());

	Scene  *scene=(Scene*)(args->self);
	Object *obj=(Object*)(args->obj);

	// if the object doesn't currently have a subwindow, show the panel and create a scene for the object
	if (!(obj->m_pPanel->visible())) {
		obj->m_pPanel->show();
		aniUI->addPanel(obj->m_pPanel);
	}
}

void Scene::cb_toggle_object(Fl_Widget* o, void* v)
{
	scene_params* args=(scene_params *)(o->user_data());

	Scene  *scene=(Scene*)(args->self);
	Object *obj=(Object*)(args->obj);

	obj->toggleVisibility();
	scene->refresh();
}

void Scene::cb_activate_object(Fl_Widget* o, void* v)
{
	scene_params* args=(scene_params *)(o->user_data());

	Scene  *scene=(Scene*)(args->self);
	Object *obj=(Object*)(args->obj);

	scene->m_pActiveObj=obj;

	aniUI->m_pCurveEditor->resetCurveMenu(scene->m_pActiveObj);
}
